#ifndef H_KNN_H

#define H_KNN_H

typedef float	mrReal;

#ifndef TRUE
#define TRUE (1)
#endif
#ifndef FALSE
#define FALSE (0)
#endif

#ifndef MIN
#define MIN(x, y)	(((x) < (y)) ? (x) : (y))
#endif

#ifndef MAX
#define MAX(x, y)	(((x) > (y)) ? (x) : (y))
#endif

#define M_DO_LOO	(1)
#define M_ADD_AVG_VAR	(1 << 1)

/*
 * The next types and statics describe the types of loss functions
 * supported. It is important to keep them in sync
 */
#define	NUM_LOSS (2)	/* Number of Loss Functions */
typedef enum lossType {L_ERR,  L_PROB}  LOSS_TYPE;
static char *lossDesc[] = {"Zero-One Loss", "Squared Prob. loss"};

typedef struct s_desc		/* Problem description */
{
  int		len_vec;	/* Dimension of the vectors */
  int		kVal;		/* K - Value of classifier */
  int		bestK[NUM_LOSS];/* Best K -Value found from CV */
  FILE		*fpPred[NUM_LOSS];/* Report Predictions */

  int		leaveOut;	/* Training template to leave out (eg for CV)*/
  int		numTempl;	/* Number of templates */
  int		totTempl;	/* Total # of templates */
  int		nClass;		/* Number of classes */
  struct Template *templates;   /* The templates */
  mrReal	machTol;	/* Machine Tolerance */

  int		nTest;		/* Number of test Points */
  mrReal	*testVec;	/* Current test Vectors */
  int		*testOut;	/* Current test output class */
  int		*pred;		/* Predictions for current test Case for different k */
  mrReal	**predProb;	/* Class probabilities */
  mrReal	distTol;	/* Tolerance for determining k nearest neighbours */
  mrReal	reportVal[NUM_LOSS];/* Values to report for loss functions */
  mrReal	*testErr[NUM_LOSS];

  int		methodOpts;	/* Method Options */
  int		verbose;
} DESC;


typedef struct Template 
{
  int		id;
  mrReal	*vec;		/* Input Vector associated with this template */
  int		out;		/* Class of this example */
  mrReal	dist2;		/* dist2 to current test example */
} TEMPLATE ;


#define STATIC_BUFFER(buffer, type, request, old)			\
	if (   ((request) > (old))					\
	    || ((type*)NULL == (buffer)))				\
	{								\
	    if ((request) > 0)						\
	    {								\
	    	if ((type*)NULL != (buffer))			\
		{							\
		    buffer = (type*)realloc((void *)(buffer), 		\
					(size_t)( (request) *sizeof(type))); \
		    memset((void *)(buffer+(old)), 0, 			\
			   (size_t)(((request)-(old))*sizeof(type)));\
		}							\
		else							\
		{							\
		    buffer = (type*)malloc((size_t)((request)* sizeof(type)));\
		    memset((void *)(buffer), 0, (size_t)((request)*sizeof(type)));\
		}							\
		(old) = (request);					\
	    }								\
	}

extern int main(int argc, char *argv[]) ;
extern void allocations( DESC *desc /* in: Problem description */ ) ;
extern void freeAllocations( DESC *desc /* in: Problem description */ ) ;
extern void get_templates( FILE *fp, DESC *desc /* in: Problem description */ ) ;
extern void testExs( DESC *desc, /* in: Problem description */ FILE *fp, /* in: Where to read test data */ int leaveOut ) ;
extern void looCv( DESC *desc, /* in: Problem description */ FILE *fp /* in: Source of the training data */ ) ;
extern int * classifyEx( DESC *desc /* in: Problem description */ ) ;
extern int evalPred( DESC *desc, /* in: Problem description */ int k /* in: k-value */ ) ;
extern int smallerDist( DESC *desc, /* in: Problem description */ int maxK, /* in Max k to consider (includes p that may be    within tol of each other (See evalPred above) */ int m, /* in: Class 1 */ int n /* Class 2 */ ) ;
extern void calc_dist2_to_one( mrReal *tst, /* in: Example Vector */ TEMPLATE *templ, /* out: Set the distance function */ int len ) ;
extern void sort_dist2( TEMPLATE  *templates, int num ) ;
extern int compare_templ( const void *arg1, const void *arg2 ) ;
extern mrReal *readRow( FILE *fp, /* in: Where to read from */ mrReal *vec, /* i/o: where to put data */ int num, /* in: Number to read */ int *out /* out: Class of example */ ) ;
extern int countFileColsAndRows( char *file, FILE *fp, /* in: Data file */ int *ncol, /* out: # of cols */ int *nlines /* out Estimated # lines */ ) ;
static int countCols( char *line, /* in: Line with data */ int maxLen /* in: Maximum line length */ ) ;
extern mrReal machineEpsilon(void) ;
static int countClasses( FILE *fp ) ;

#endif
